/*
**
**  Surface list
**  handling of Keyboard events
**  Some subtilties of surface list flags  (autodowndate, autoupdateexpose)
**  Loading jpeg graphics
**  writing a screenshot 
**  use timing functions
*/ 



/*
we create Two surfaces:
   the left one displays a jpeg which is changed with the keyboard. It needs to be
   displayed only when a new image is loaded. So the flag 'UPDATED' need to be
   set only at this time, then cleared once it is displayed. 
   It could be cleared by hand, but I prefered to use the flag 'AUTODOWNDATE' which
   clears the UPDATED flag each time the image is displayed.
   This image needs however to be redrawn if the window get obscured then visible again.
   So I set the flag 'AUTOUPDATEEXPOSE' which force a redraw by setting  UPDATED on
   such events. (note this is a workaround before AUTOREDRAW is implemented)
   
   The right one displays a constantly changing image, so 'UPDATED' is set all time.
   You can control if the image is displayed with the keaybord; the flags UPDATED
   is then toggled (by the program).

Timing:
   the right image much move at the same speed on all machines, so I use an ot_time
   structure. It serves to record the time of program start and the current time,
   and associated functions return you a parameter (by scaling the elapsed time)
   that you can use as a basis for the modification of images.

Loading of graphics:
   The program loads jpeg files with ofn_JpegLoad(), then transfer them in the left image
   to be displayed.

Screenshot:
   the function ofn_SurfaceListSnapshot() writes the surfaces of the list which conform
   to the mask provided (usually it's simply those that are VISIBLE) in a ot_simple_image.
   This image can then be written on disk with ofn_JpegWrite().

Events:
   To get keyboard events and perform automatic redraw of the left image on X Exposure events,
   we use ofn_ProcessSurfaceListEvents().
   This function does internally what it can, then return you the remaining (keyboard) events
   one by one. That is, you must call it in loop until it doesn't return anything.
 
   The keyboard codes it returns are the same than the X keysyms .
 
 
  

*/


   








#include <stdio.h>
#include <math.h>

#include "o_std.h"
#include "o_types.h"
#include "o_xutils.h"
#include "o_surface.h"
#include "o_gfxm.h"

#include "o_gfxfile.h"  // to load and write jpegs
#include "o_handler.h"  // for events
#include "o_schedule.h"  // for time

#define WIDTH  320
#define HEIGHT 240
#define W1 100
#define W2 220


char *jfile[4] = {"ex4-jfile1.jpg","ex4-jfile2.jpg","ex4-jfile3.jpg","ex4-jfile4.jpg"};



void Rotate8(ot_surface *OS8, double angle){
  uint8 *tmpbuffer;
  double fx,fy;
  int x,y;


  fx = cos(angle);
  fy = sin(angle);
  tmpbuffer = (uint8 *)OS8->buffer;
   for (y = 0; y< HEIGHT; y++)
     for (x = 0; x< W2; x++)
       {
	 *(tmpbuffer + x + y*W2) = floor(x*fx+y*fy);

       }
}




void Load32(ot_surface *OS32, char *filename){
  ot_simple_surface *image;

  image = ofn_JpegLoad( filename, OS32->gfxMode);
  if (image == NULL){
    printf("%s : error loading file\n", filename);
    return;
  }
  if ( (image->width != OS32->width)
       || (image->height != OS32->height)){
    printf("%s : Wrong image dimensions \n",filename);
  } else {
    memcpy(OS32->buffer, image->buffer,
	   OS32->width * OS32->height * ofn_GetBppOfMode(OS32->gfxMode)/8);
  }

  ofn_SimpleSurfaceDestroy(image);
  free(image);

  
}


void usage(){
  printf("\n\n");
  printf(" Try these key \n");
  printf(" left/right : cycle through the jpeg files displayed on the left side\n");
  printf(" space      : stop/restart the right side \n");
  printf(" s          : save a screenshot \n");
  printf(" q/esc      : quit \n");
  printf("\n\n");
   
}

void main(){
  ot_display      *ODis;
  ot_top_window   *OWin;
  ot_surface      *OSurf[2];
  int              index[2];  // handle of the surfaces in the list;
  ot_surface_list *OSList;
  ot_color        *palette;
  int              i;
  ot_time         *OTime;
  int              jpeg_index = 0;

  // allocate these structures. 

  OTime   = NEW(ot_time);
  ODis = NEW(ot_display);
  OWin = NEW(ot_top_window);
  OSurf[0]= NEW(ot_surface);
  OSurf[1]= NEW(ot_surface);
  OSList  = NEW(ot_surface_list);
  palette = NEW_ARRAY(ot_color,256);


  // Open display and create a window

  ofn_XOpen(ODis, "",True);
  ofn_XOpenWindow(ODis, OWin, WIDTH, HEIGHT, "Example4 : Events, files, timer");

  // see what Optilib has detected
  printf(" XShm extension : %s\n", 
	 ODis->o_XShmAvailable ? "In use": "Not in use");
  printf(" Display mode : %s\n",
	 ofn_GetNameOfMode(ODis->gfxMode));
  if (ODis->gfxMode == MODE_PAL_8)
    printf(" Warning, your 8bpp server will only display trash.... sorry...\n");
  


  usage();

  // set the palette for the first surface
  for (i=0; i<256; i++){
    palette[i].r = (int)(i*0.7);
    palette[i].g = (int)(i*0.8);
    palette[i].b = (i)&255;
  } 

  // Create a 8bpp surface and attach it to the X window
   ofn_SurfaceCreate( OSurf[0], W2, HEIGHT,  MODE_PAL_8);
   ofn_SurfaceAttachX( OSurf[0], OWin, W1,0);
   ofn_SurfaceSetPalette(OSurf[0], palette);

   // you can now free the palette
   free(palette);


  // Create a truecolor image

   // If you want, you can create an image in the same colormode as the display
   // to avoid conversion (and RAM). In this case this is simple as we only load a jpeg
   // in it; but in other case you have to write customary routines ( and this
   // is what Optilib try to avoid ;) )
   if (ODis->gfxMode != MODE_PAL_8)
     ofn_SurfaceCreate( OSurf[1], W1, HEIGHT,  ODis->gfxMode);
   else
     ofn_SurfaceCreate( OSurf[1], W1, HEIGHT,  MODE_RGB_565);
   ofn_SurfaceAttachX( OSurf[1], OWin, 0,0); 
   Load32( OSurf[1], jfile[0]);


   index[0] = ofn_SurfaceListAddElement(OSList, OSurf[0], OPL_SLIST_VISIBLE | OPL_SLIST_UPDATED);
   index[1] = ofn_SurfaceListAddElement(OSList, OSurf[1], OPL_SLIST_VISIBLE | OPL_SLIST_UPDATED
 					                | OPL_SLIST_AUTODOWNDATE | OPL_SLIST_AUTOUPDATEEXPOSE);




   // get start time, initialize things
   ofn_sched_init(OTime);


   /* Now this is the main loop */
   for (;;){
     ot_event oev;
     int stop = 0;

     // get current time, compute difference since start, etc
     ofn_sched_update(OTime);

     
     // it's useless to compute it if it's not displayed ;)
     if (OSList->flags[index[0]] & OPL_SLIST_UPDATED) 
       Rotate8(OSList->list[index[0]],
	       ofn_scale_time( 0.5, &(OTime->demo_elapsed_time))
	       ); 
	       

     ofn_SurfaceListBlit(OSList, OPL_SLIST_VISIBLE | OPL_SLIST_UPDATED);
     do{
       ofn_ProcessSurfaceListEvents(OSList, &oev);
       if ( (oev.type == OPL_EV_KEY) && (oev.ev_key.pressed))
	 switch(oev.ev_key.keysym){
	 case XK_Escape:   
	 case XK_q:   
	 case XK_Q:   
	   stop = 1;
	   break;
	 case XK_Left:
	 case XK_KP_Left:
	   jpeg_index--;
	   Load32( OSList->list[index[1]], jfile[jpeg_index & 3]);
	   OSList->flags[index[1]] |= OPL_SLIST_UPDATED; 
	   break;

	  case XK_Right:
	  case XK_KP_Right:
	    jpeg_index++;
	    Load32( OSList->list[index[1]], jfile[jpeg_index & 3]);
	    OSList->flags[index[1]] |= OPL_SLIST_UPDATED; 
	    break;

	 case XK_space:  // toggle right image
	    OSList->flags[index[0]] ^= OPL_SLIST_UPDATED; 
	    break;

	  case XK_s:
	    {
	      ot_simple_surface *snap;
	      snap = ofn_SurfaceListSnapshot (OSList, MODE_ARGB_8888, OPL_SLIST_VISIBLE);
	      printf(" Writing screenshot %s \n","ex4-screenshot.jpg");
	      ofn_JpegWrite("ex4-screenshot.jpg",snap);
	      ofn_SimpleSurfaceDestroy(snap);
	      free(snap);
	      
	    }   
	    break;
	 default:
	   break;
	 }

     } while (oev.type != OPL_EV_NOTHING);


     if (stop)
       break;
   }

   // Release structures
   ofn_SurfaceListDestroyAll(OSList); 
   free(OSList); 
   /* ofn_SurfaceListDestroyAll detach, destroy and free the
    * associated surfaces.
    */

   
   ofn_XCloseWindow(OWin);
   ofn_XClose(ODis);

   free(OWin);
   free(ODis);
   free(OTime);
     

}







